// Copyright (c) The go-grpc-middleware Authors.
// Licensed under the Apache License 2.0.

/*
Package prometheus provides a standalone interceptor for metrics. It's next iteration of deprecated https://github.com/grpc-ecosystem/go-grpc-prometheus.
See https://github.com/grpc-ecosystem/go-grpc-middleware/tree/main/examples for example.
*/
package prometheus
