// Copyright (c) 2016, Daniel Martí <mvdan@mvdan.cc>
// See LICENSE for licensing information

// Package syntax implements parsing and formatting of shell programs.
// It supports POSIX Shell, Bash, and mksh.
package syntax
