// Code generated by generate.go. DO NOT EDIT.

package simple

import (
	"honnef.co/go/tools/analysis/lint"
	"honnef.co/go/tools/simple/s1000"
	"honnef.co/go/tools/simple/s1001"
	"honnef.co/go/tools/simple/s1002"
	"honnef.co/go/tools/simple/s1003"
	"honnef.co/go/tools/simple/s1004"
	"honnef.co/go/tools/simple/s1005"
	"honnef.co/go/tools/simple/s1006"
	"honnef.co/go/tools/simple/s1007"
	"honnef.co/go/tools/simple/s1008"
	"honnef.co/go/tools/simple/s1009"
	"honnef.co/go/tools/simple/s1010"
	"honnef.co/go/tools/simple/s1011"
	"honnef.co/go/tools/simple/s1012"
	"honnef.co/go/tools/simple/s1016"
	"honnef.co/go/tools/simple/s1017"
	"honnef.co/go/tools/simple/s1018"
	"honnef.co/go/tools/simple/s1019"
	"honnef.co/go/tools/simple/s1020"
	"honnef.co/go/tools/simple/s1021"
	"honnef.co/go/tools/simple/s1023"
	"honnef.co/go/tools/simple/s1024"
	"honnef.co/go/tools/simple/s1025"
	"honnef.co/go/tools/simple/s1028"
	"honnef.co/go/tools/simple/s1029"
	"honnef.co/go/tools/simple/s1030"
	"honnef.co/go/tools/simple/s1031"
	"honnef.co/go/tools/simple/s1032"
	"honnef.co/go/tools/simple/s1033"
	"honnef.co/go/tools/simple/s1034"
	"honnef.co/go/tools/simple/s1035"
	"honnef.co/go/tools/simple/s1036"
	"honnef.co/go/tools/simple/s1037"
	"honnef.co/go/tools/simple/s1038"
	"honnef.co/go/tools/simple/s1039"
	"honnef.co/go/tools/simple/s1040"
)

var Analyzers = []*lint.Analyzer{
	s1000.SCAnalyzer,
	s1001.SCAnalyzer,
	s1002.SCAnalyzer,
	s1003.SCAnalyzer,
	s1004.SCAnalyzer,
	s1005.SCAnalyzer,
	s1006.SCAnalyzer,
	s1007.SCAnalyzer,
	s1008.SCAnalyzer,
	s1009.SCAnalyzer,
	s1010.SCAnalyzer,
	s1011.SCAnalyzer,
	s1012.SCAnalyzer,
	s1016.SCAnalyzer,
	s1017.SCAnalyzer,
	s1018.SCAnalyzer,
	s1019.SCAnalyzer,
	s1020.SCAnalyzer,
	s1021.SCAnalyzer,
	s1023.SCAnalyzer,
	s1024.SCAnalyzer,
	s1025.SCAnalyzer,
	s1028.SCAnalyzer,
	s1029.SCAnalyzer,
	s1030.SCAnalyzer,
	s1031.SCAnalyzer,
	s1032.SCAnalyzer,
	s1033.SCAnalyzer,
	s1034.SCAnalyzer,
	s1035.SCAnalyzer,
	s1036.SCAnalyzer,
	s1037.SCAnalyzer,
	s1038.SCAnalyzer,
	s1039.SCAnalyzer,
	s1040.SCAnalyzer,
}
