// Code generated by "stringer -type=quoteState"; DO NOT EDIT.

package syntax

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[noState-1]
	_ = x[subCmd-2]
	_ = x[subCmdBckquo-4]
	_ = x[dblQuotes-8]
	_ = x[hdocWord-16]
	_ = x[hdocBody-32]
	_ = x[hdocBodyTabs-64]
	_ = x[arithmExpr-128]
	_ = x[arithmExprLet-256]
	_ = x[arithmExprCmd-512]
	_ = x[arithmExprBrack-1024]
	_ = x[testExpr-2048]
	_ = x[testExprRegexp-4096]
	_ = x[switchCase-8192]
	_ = x[paramExpName-16384]
	_ = x[paramExpSlice-32768]
	_ = x[paramExpRepl-65536]
	_ = x[paramExpExp-131072]
	_ = x[arrayElems-262144]
}

const _quoteState_name = "noStatesubCmdsubCmdBckquodblQuoteshdocWordhdocBodyhdocBodyTabsarithmExprarithmExprLetarithmExprCmdarithmExprBracktestExprtestExprRegexpswitchCaseparamExpNameparamExpSliceparamExpReplparamExpExparrayElems"

var _quoteState_map = map[quoteState]string{
	1:      _quoteState_name[0:7],
	2:      _quoteState_name[7:13],
	4:      _quoteState_name[13:25],
	8:      _quoteState_name[25:34],
	16:     _quoteState_name[34:42],
	32:     _quoteState_name[42:50],
	64:     _quoteState_name[50:62],
	128:    _quoteState_name[62:72],
	256:    _quoteState_name[72:85],
	512:    _quoteState_name[85:98],
	1024:   _quoteState_name[98:113],
	2048:   _quoteState_name[113:121],
	4096:   _quoteState_name[121:135],
	8192:   _quoteState_name[135:145],
	16384:  _quoteState_name[145:157],
	32768:  _quoteState_name[157:170],
	65536:  _quoteState_name[170:182],
	131072: _quoteState_name[182:193],
	262144: _quoteState_name[193:203],
}

func (i quoteState) String() string {
	if str, ok := _quoteState_map[i]; ok {
		return str
	}
	return "quoteState(" + strconv.FormatInt(int64(i), 10) + ")"
}
